#!/usr/bin/perl
#############################################################################
#
# Impostor class IRC Robot
# $Id: impostor.pl,v 1.20 2004/06/06 14:07:26 cade Exp $
# Vladimir Ivanov Belperchinov-Shabanski <cade@biscom.net> 2004
# DISTRIBUTED UNDER GNU GPL LICENSE, SEE `COPYING' FILE OR WWW.FSF.ORG
#
# initially made for Wolfenstein: Enemy Territory services at
# http://play.evrocom.net
#
#############################################################################
use Time::HiRes qw ( sleep );
use IO::Socket::INET;
use IO::Select;
use IO::File;
use Fcntl;
use Data::Dumper;
use Net::IRC;
use strict;

our $DEBUG = 0 || $ENV{'DEBUG'};

### GENERAL CONFIG ##########################################################

our $NICK = 'amee';          # nick
our $HERE = $ENV{'IMPOSTOR_HERE'} || 'play.evrocom.net';  # bind address of the host machine

our %NETWORKS = ( # still not in use
                );

# which servers to connect and which channels to join
our %SERVERS  = ( 'irc.evrocom.net'  => [ '#et', '#warez.ru', '#play', '#play1', '#play2', '#yinyen' ],
                  'irc.spnet.net'    => [ '#et', '#[LUD]' ],
                  'irc.freenode.net' => [ '#ets' ],
 );

# for debug mode set just one server and channel
our $NICK     = 'impostor_hidden' if $DEBUG;
our %SERVERS  = ( 'irc.mtel.net' => [ '#dm' ] ) if $DEBUG;

our %IRC; # do not touch this one, holds server-name => connection-handler map

our $RECONNECT_TIME = 60; # seconds to wait before reconnect


### Wolfenstein: Enemy Territory stuff ######################################

# this config is for queries: `!ets server'
our %ETSTAT_SERVERS = ( # nick => address
                      'play'   => 'play.evrocom.net',
                      'play1'  => 'play.evrocom.net',
                      'play2'  => 'play.evrocom.net:27961',
                      'online' => 'games.online.bg',
                      'tjw'    => 'et.tjw.org',
                      );
# rcon passwords, required for !rcon and chat
our %ETRCON_PASSWD  = ( # nick => rcon password
                      'play'    => 'password1',
                      'play1'   => 'password2',
                      'play2'   => 'password3',
                      );

# W:ET services, usualy you don't need to touch those
our %INQUIRIES = (
                 # !rcon ET access:
                 # !rcon ident password1
                 # !rcon server rcon-command ...
                 # <server> is one of %ETSTAT_SERVERS, but just ones with
                 # password in %ETRCON_PASSWD
                 '!rcon' => {                       # inquire id
                            IDENT => 'password1',   # password to ident
                            IDENT_EXPIRE => 120*60, # ident expire
                            HANDLER => \&i_rcon,    # handler
                            },
                 # ET name resolver:
                 # !etr ident password
                 # !etr nick
                 # <nick> can be partial
                 # comment two IDENT lines below to get password free access
                 # to !etr
                 '!etr' =>  {
                            IDENT => 'password2',
                            IDENT_EXPIRE => 20*60,
                            HANDLER => \&i_etresolve,
                            },
                 # this is replaced by transcribtions, check above
                 # otherwise you can send message with:
                 # !etchat <server> message           
                 # <server> is one of %ETSTAT_SERVERS, but just ones with
                 # password in %ETRCON_PASSWD
                 '!etchat' =>  {
                            HANDLER => \&i_etchat,
                            },
                 # !ets server
                 # <server> is one of %ETSTAT_SERVERS, but just ones with
                 # password in %ETRCON_PASSWD
                 # this will print current players
                 '!ets' =>  { 
                            # this is ident-less inqire 
                            HANDLER => \&i_etstat, # just handler
                            }
                 );

# if you want to ET players via IRC, you have to specify channels here
# then you can join those channels and talk with ET players with `nick: msg'
# (nick is this bot's nick, see above $NICK
our %PUB_HANDLERS = (
                    '#play1' => {
                                TO => 1, # answer only if addressed as `to: msg'
                                SERVER => 'play', # check %ETSTAT_SERVERS above
                                HANDLER => \&i_etchat2, # handler, don't touch pls
                                },
                    '#play2' => {
                                TO => 1, # answer only if addressed as `to: msg'
                                SERVER => 'play2', # check %ETSTAT_SERVERS above
                                HANDLER => \&i_etchat2, # handler, don't touch pls
                                }
                    );

# flood control
our $ETCHAT_FLOOD_TIME  = 60; # message foold time window, 1 minute
our $ETCHAT_FLOOD_COUNT = 10; # message count per $ETCHAT_FLOOD_TIME window
our %ETCHAT_FLOOD_USER  = (); # hash with flood arrays per user

# transcribtion config, ET to IRC chat pipe
our %LOOP_HANDLERS = (
                     'et_transcriber1' => {
                                # which IRC server
                                SERVER => 'irc.evrocom.net', 
                                # which IRC channel
                                CHAN => '#play1',
                                # which ET server, check %ETSTAT_SERVERS above
                                ET_SERVER => 'play1',
                                # wait time between loops, in seconds
                                TIME => 10,
                                # log file, note: this is console log, not games log!
                                FILE => '/home/furiozo/.etwolf/etmain/etconsole.log',
                                # do not touch this handler pls
                                HANDLER => \&l_et_transcriber, # handler
                                },
                     'et_transcriber2' => {
                                # same as above but for different server/channel
                                SERVER => 'irc.evrocom.net',
                                CHAN => '#play2',
                                ET_SERVER => 'play2',
                                TIME => 10,
                                FILE => '/home/furiozo/.etwolf-pub/etmain/etconsole.log',
                                HANDLER => \&l_et_transcriber, # handler
                                },
                     );
                     
our %ET_GUID_REF = (
                   #---cade--------------------------------------------------
                   'GUID:447346534653456345634563465563dc' => 'cade',
                   'IP:192.168.1.14' => 'cade',
                   #---------------------------------------------------------
                   );
                    
our %ET_GUID_REF_CMDS = (
                        '!?' => 'say I\'m here, go ahead',
                        '!pause' => 'ref pause',
                        '!unpause' => 'ref unpause',
                        '!lock' => 'ref lock',
                        '!unlock' => 'ref unlock',
                        '!cancelvote' => 'cancelvote',
                        '!passvote' => 'passvote',
                        );

# path to qstat utility
# this won't be required in future releases, it is still in use due to history reasons
# sorry
# here is the qstat home page:
# http://www.qstat.org/
my $QSTAT = "/home/cade/bin/qstat"; 

# this file holds info about nicks and guids
# please check `make-resolve-data.sh' tool from this package!
my $GUID_FILE = "/home/cade/guid.txt.gz";

### MAIN, GO! ###############################################################
#
# *** you don't need to fix anything below this line, thanks ***
#

our %INQ_IDENT; # this holds inquire+full-hostmask which are ident'd

our %RECONN_TIME = (); # holds reconnect time for each server

$| = 1;

my $break_main_loop = 0;

$SIG{ 'INT'  } = sub { $break_main_loop = 1; };
$SIG{ 'HUP'  } = sub { $break_main_loop = 1; };   
$SIG{ 'TERM' } = sub { $break_main_loop = 1; };   

my $irc_generic = new Net::IRC;

irc_connect_server( $_ ) for( keys %SERVERS );

while(4)
  {
  $irc_generic->do_one_loop();
  while( my ( $server, $irc ) = each %IRC )
    {
    if( $irc->connected() )
      {
      # server is connected
      $irc->out_flush1(); # print one line to each user
      next;
      }
    # server disconnected
    my $rct = $RECONN_TIME{ $server };
    if( ! $rct )
      {
      # there is no reconnect time
      $RECONN_TIME{ $server } = time() + $RECONNECT_TIME;
      # set reconnect time in the future
      logger( $irc, "RECONNECT: delay $RECONNECT_TIME seconds" );
      next;
      }
    else
      {
      # check if reconnect time is now or past
      next if $rct > time();
      }
    # reconnect!  
    $RECONN_TIME{ $server } = 0;
    $irc->disconnect(); # just to be safe :)
    logger( $irc, "RECONNECT" );
    irc_connect_server( $server );
    }
  if( 1 ) #time() % 10 == 0 )
    {
    for my $l ( keys %LOOP_HANDLERS )
      {
      my $server = $LOOP_HANDLERS{ $l }{ 'SERVER' };
      my $irc = $IRC{ $server } or next;
      next if time() < $LOOP_HANDLERS{ $l }{ 'LOOP_TIME' };
      $LOOP_HANDLERS{ $l }{ 'LOOP_TIME' } = time() + $LOOP_HANDLERS{ $l }{ 'TIME' };
      my $h = $LOOP_HANDLERS{ $l }{ 'HANDLER' };
      if( ! $h )
        {
        logger( $irc, "FATAL: no LOOP_*:HANDLER for [$l]" );
        }
      else
        {
        &$h( $irc, $l );
        }
      }
    };
  last if $break_main_loop; # signals controlled
  }
logger( undef, "EXIT\n" );

### FUNCTIONS ###############################################################

sub irc_connect_server
{
  my $server = shift;
  
  my $irc = $irc_generic->newconn( Nick      => $NICK,
                                   Server    => $server,
                                   Port      =>  6667,
                                   Ircname   => $NICK,
                                   Username  => $NICK,
                                   LocalAddr => $HERE );
  if( ! $irc )
    {
    logger( undef, "ECONNECT: $server" );
    return 0;
    }

  logger( $irc, "SERVER: $server" );
  
  $irc->add_handler( 'msg', \&on_msg);
  $irc->add_handler( 'public', \&on_public);
  $irc->add_handler( 'chat', \&on_dcc_chat);
  $irc->add_handler( 'dcc_open', \&on_dcc_chat);
  $irc->add_handler( 'dcc_update', \&on_dcc_chat);
  $irc->add_handler( 'dcc_close', \&on_dcc_chat);
  $irc->add_handler( 'cdcc', \&on_cdcc);
  $irc->add_handler( 'cping', \&on_ctcp_ping);
  $irc->add_handler( 'cversion', \&on_ctcp_version);
  
  $irc->join( $_ ) and logger( $irc, "JOIN: $_" ) for ( @{ $SERVERS{ $server } } );
  
  $IRC{ $server } = $irc;
  $RECONN_TIME{ $server } = 0;
  return 1;
}

### HANDLERS ################################################################

sub on_msg
{
  $_[2] = 1;
  $_[1]->{ '_PRIVATE' } = 1; # set event to private
  on_privmsg( @_ );
}

sub on_public
{
  $_[2] = 0;
  $_[1]->{ '_PRIVATE' } = 0; # set event to public
  on_privmsg( @_ );
}

sub on_dcc_chat
{
  # FIXME: ...
  print Dumper( @_ );
}

sub on_cdcc
{
  my $self = shift;
  my $ev   = shift;
  
  my $f = $ev->from();
  my $n = $ev->nick();  
  my @args = $ev->args();
  my $args = join ' ', @args;

  if( $args =~ /^CHAT/ )
    {
    # received DCC CHAT request
    $self->new_chat( $ev );
    logger( $self, "*** DCC CHAT: from $f received" );
    $self->privmsg( $n, "*** DCC CHAT accepted" );
    }
}

sub on_privmsg
{
  my $self = shift;
  my $ev   = shift;
  my $prv  = shift; # private
  
  my $f = $ev->from();
  my $t = ($ev->to())[0];
  my $n = $ev->nick();  
  
  my @args = $ev->args();
  my $args = join ' ', @args;
  # print "*** PRIVMSG: $n ($f) -> $t [@args]\n";

  if( $args =~ /^\s*(!\S+)(.*)$/ )
    {
    # !inquire received
    my $res = process_inquire( $self, $ev, $1, $2 );
    # if private return to user, otherwise to the channel
    $self->out( $ev->{ '_PRIVATE' } ? $n : $t, $res ) if $res ne '';
    return;
    }
  if( ! $ev->{ '_PRIVATE' } and $PUB_HANDLERS{ $t } )
    {
    my $h = $PUB_HANDLERS{ $t }{ 'HANDLER' };
    if( ! $h )
      {
      logger( $self, "FATAL: no PUB_*:HANDLER for [$t]" );
      }
    else
      {
      # print "DEBUG: calling inquire handler $h for $inq\n";
      }
    
    return if $PUB_HANDLERS{ $t }{ 'TO' } and $args !~ s/^$NICK\s*:\s*//i;
    
    $args =~ s/^\s+//;
    $args =~ s/\s+$//;
    my $res = &$h( $self, $ev, $args ); # call public message handler!
    return;
    }  
}

sub process_inquire
{
  my $self =    shift;
  my $ev   =    shift;
  my $inq  = lc shift;
  my $cmd  =    shift;
  
  my $f = $ev->from();
  my $t = ($ev->to())[0];
  my $n = $ev->nick();  
  
  if( ! $INQUIRIES{ $inq } )
    {
    # we don't know for such inquire
    logger( $self, "ERROR: invalid inqire [$inq] received from $f" );
    return;
    }

  if( $INQUIRIES{ $inq }{ 'IDENT' } )
    {
    # this inquire requires ident password
    logger( $self, "IDENT CHECK: inquire [$inq] received from $f" );
    if( $cmd =~ /^\s*ident\s+(\S+)/i )
      {
      # user tries to ident...
      if( $1 eq $INQUIRIES{ $inq }{ 'IDENT' } )
        {
        # ident password is ok, set expire time in the future
        $INQ_IDENT{ $inq }{ $f } = time() + $INQUIRIES{ $inq }{ 'IDENT_EXPIRE' };
        return "*** $inq: identification ok ($f)";
        logger( $self, "IDENT CHECK: OK: inquire [$inq] received from $f" );
        }
      else
        {
        # ident password is wrong, remove ident information
        $INQ_IDENT{ $inq }{ $f } = 0;
        return "*** $inq: identification failed";
        logger( $self, "IDENT CHECK: FAILED: inquire [$inq] received from $f" );
        }  
      }
    # check if user has ident or if ident is expired  
    if( $INQ_IDENT{ $inq }{ $f } and time() > $INQ_IDENT{ $inq }{ $f } )
      {
      # expired!
      $INQ_IDENT{ $inq }{ $f } = 0;
      return "*** $inq: error: your ident expired, identify again";
      }
    # check if user has ident ok  
    return "*** $inq: error: identify first: $inq ident <password>" unless $INQ_IDENT{ $inq }{ $f };
    # ident is ok, (re)set expire time in the future
    $INQ_IDENT{ $inq }{ $f } = time() + $INQUIRIES{ $inq }{ 'IDENT_EXPIRE' };
    }
    
  logger( $self, "OK: inquire [$inq] received from $f" );
  my $h = $INQUIRIES{ $inq }{ 'HANDLER' };
  if( ! $h )
    {
    logger( $self, "FATAL: no HANDLER for [!inq]" );
    return "*** $inq: error: internal, try again later";
    }
  else
    {
    # print "DEBUG: calling inquire handler $h for $inq\n";
    }  
  $cmd =~ s/^\s+//;
  $cmd =~ s/\s+$//;
  my $res = &$h( $self, $ev, $cmd ); # call inqire handler!
  # return to user or channel regarding PRIVATE mark
  $self->out( $ev->{ '_PRIVATE' } ? $n : $t, $res ) if $res ne '';
}

### OVERRIDE! ###############################################################

package Net::IRC::Connection;

# to be able to send information to many users simultaneously we should have
# buffer which holds information for every user and return one line at a time
# to each one!

sub out
{
  my $self = shift; # connection object
  my $who  = shift; # recipient
  my @msg  = @_;    # message(s)
  
  # create new buffer if don't exist...
  $self->{ '_OUTPUT' }{ $who } ||= [];
  
  for my $l ( @msg )
    {
    for( split /\n/, $l )
      {
      next unless $_;
      push @{ $self->{ '_OUTPUT' }{ $who } }, $_;
      }
    }
}

# this prints back one line for each active recipient
sub out_flush1
{
  my $self = shift;
  # make empty hash for output buffer if don't exist
  $self->{ '_OUTPUT' } ||= {};
  for my $who ( keys %{ $self->{ '_OUTPUT' } } ) # loop on recipients
    {
    # remove buffer to recipient if it is empty
    delete $self->{ '_OUTPUT' }{ $who } and next unless @{ $self->{ '_OUTPUT' }{ $who } };
    # get next line
    my $line = shift @{ $self->{ '_OUTPUT' }{ $who } };
    # print "[$who:$line]";
    next if $line eq '';
    # send the line
    $self->privmsg( $who, $line );
    }
}

package main;

### CTCP REPLIES ############################################################

sub on_ctcp_ping 
{
  my $self = shift;
  my $ev   = shift;

  my $n = $ev->nick();  

  $self->ctcp_reply( $n, 'PING ' . join ( ' ', ( $ev->args() ) ) );
  logger( $self, "*** CTCP PING request from [$n] received" );
}

sub on_ctcp_version
{
  my $self = shift;
  my $ev   = shift;

  my $n = $ev->nick();  

  $self->ctcp_reply( $n, 'VERSION Impostor class robot rev. $Revision: 1.20 $' );
  logger( $self, "*** CTCP VERSION request from [$n] received" );
}

### UTILITIES ###############################################################

sub logger
{
  my $irc = shift;
  my $msg = shift;
  
  my $srv = ref($irc) eq 'Net::IRC::Connection' ? $irc->server() : 'GLOBAL';
  my $tm = scalar localtime();
  print "[$tm] ($srv) $msg\n";
}

### INQUIRIES HANDLERS ######################################################

sub i_test
{
  my $self =    shift;
  return "*** hello, i_test here ***";
}

#----------------------------------------------------------------------------

sub i_etchat
{
  my $self = shift;
  my $ev   = shift;
  my $cmd  = shift;

  my $n = $ev->nick();  

  $self->notice( $n, "!etchat is not available, talk to me directly" );
  return;
  
  return "*** usage: !etchat server message" unless $cmd =~ /^(\S+)\s+(.+)$/i;
  return "*** unknown server [$1]" unless $ETSTAT_SERVERS{ $1 };
  et_rcon( $1, "say $n: $2" ); # no reasonable output here...
  return; 
}

#----------------------------------------------------------------------------

sub i_etchat2
{
  my $self = shift;
  my $ev   = shift;
  my $cmd  = shift;

  my $f = $ev->from();
  my $t = ($ev->to())[0];
  my $n = $ev->nick();  

  $ETCHAT_FLOOD_USER{ $f } ||= [];
  my $ar = $ETCHAT_FLOOD_USER{ $f };
  
  my $mc = @$ar; # message count
  if( $mc >= $ETCHAT_FLOOD_COUNT )
    {
    my $ot = $$ar[0]; # oldest message time
    my $tp = time() - $ot; # time period
    print sprintf "FLOOD: ot:%d, mc:%d, tp:%d, %.3f > %.3f\n", $ot, $mc, $tp, $mc / $tp, $ETCHAT_FLOOD_COUNT / $ETCHAT_FLOOD_TIME;
    
    if( $tp < 1 or ( $mc/$tp ) > ( $ETCHAT_FLOOD_COUNT/$ETCHAT_FLOOD_TIME ) )
      {
      my $ws = int( ( $ot + ( $mc * $ETCHAT_FLOOD_TIME / $ETCHAT_FLOOD_COUNT ) ) - time() );
      $self->notice( $n, "flood control: etchat ignored, please wait $ws seconds" );
      return;
      }
    shift @$ar;
    }
  push  @$ar, time();  

  my $server = $PUB_HANDLERS{ $t }{ 'SERVER' };
  #print "ETCHAT2: $server: say $n: $cmd\n";
  et_rcon( $server, "say $n: $cmd" ); # no reasonable output here...
  return; 
}

#----------------------------------------------------------------------------

sub i_etstat
{
  my $self = shift;
  my $ev   = shift;
  my $cmd  = shift;
  
  return "*** usage: !ets server" unless $cmd =~ /^(\S+)/i;
  return "*** unknown server [$1]" unless $ETSTAT_SERVERS{ $1 };
  return et_stat( $1 );
}

#----------------------------------------------------------------------------

sub i_rcon
{
  my $self = shift;
  my $ev   = shift;
  my $cmd  = shift;
  
  return "*** usage: !rcon server message" unless $cmd =~ /^(\S+)\s+(.+)$/i;
  return "*** unknown server [$1]" unless $ETSTAT_SERVERS{ $1 };
  return et_rcon( $1, $2 );
}

#----------------------------------------------------------------------------

sub l_et_transcriber
{
  my $self = shift;
  my $loop_id = shift;
  
  my $chan = $LOOP_HANDLERS{ $loop_id }{ 'CHAN' };
  
  my $fh = $LOOP_HANDLERS{ $loop_id }{ 'FH' };
  if( ! $fh )
    {
    my $fn = $LOOP_HANDLERS{ $loop_id }{ 'FILE' };
    $fh = new IO::File( "/usr/bin/tail -f $fn |" ) or logger( $self, "ERROR: et_transcriber: cannot open $fn ($!)" );
    my $flags = '';
    fcntl( $fh, F_GETFL, $flags ) or logger( $self, "ERROR: et_transcriber: cannot get flags $fn ($!)" );
    $flags |= O_NONBLOCK;
    fcntl( $fh, F_SETFL, $flags ) or logger( $self, "ERROR: et_transcriber: cannot set flags $fn ($!)" );
    $LOOP_HANDLERS{ $loop_id }{ 'FH' } = $fh;
    }
    
  my $c;
  while(4)
    {
    last if $break_main_loop;
    $_ = <$fh>;
    last unless defined $_;
    s/\^.//go;
    s/^([\d\:\.]+)?\s*//;
    my $ok = 0;
    $ok = 1 if ! $ok and /Kick Command Issued/i;
    $ok = 1 if ! $ok and /Setting (Axis|Allied) autospawn to/i;
    $ok = 1 if ! $ok and /^Exit:/i;
    $ok = 1 if ! $ok and /^red:/i;
    $ok = 1 if ! $ok and s/^say:\s*//i;
    $ok = 1 if ! $ok and /^score:/;
    $ok = 1 if ! $ok and s/^broadcast: print \"(.+?)\\n\"?\s*$/broadcast: $1/i; # "
    if( ! $ok and /^InitGame:\s*(.*)$/o )
      {
      my %s = split /\\/, "\\$1";
      $self->out( $chan, "********* init game **********" );
      for( qw( version mapname timelimit ) )
        {
        $self->out( $chan, "$_: " . $s{$_} );
        }
      # $ok = 1;
      next;
      }
    # $ok = 1; # DEBUG
    next unless $ok;
    $self->out( $chan, $_ );
    # print STDERR "PRE GUID: [$_]\n";
    et_guid_referee( $self, $loop_id, $1, $2 ) if /^(.+?):\s+(!.+)$/;
  }
}

sub et_guid_referee
{
  my $self    = shift;
  my $loop_id = shift;
  my $player  = shift;
  my $arg     = shift;
  
  my @arg = split /\s+/, $arg;

  my $et_server = $LOOP_HANDLERS{ $loop_id }{ 'ET_SERVER' } or return;
  
  my @rcon = split /\n/, et_rcon( $et_server, "pb_sv_plist" );
  my $found = 0;
  for( @rcon )
    {
    s/^PunkBuster Server:\s*//;
    next if /^(End of )?Player List/;
    my @pb = split /\s+/, $_, 9;
    # 0 [1]
    # 1 [f112ad6c41a48ada6b4cfc2e76fac9c1(?)]
    # 2 [193.200.15.141:29111]
    # 3 [OK]
    # 4 [1]
    # 5 [0.0]
    # 6 [0]
    # 7 [(W)]
    # 8 [[ZaZ]KserkS]
    $pb[1] =~ s/\(.*\)$//;
    $pb[2] =~ s/:\d*$//;
    if( $player eq $pb[8] and 
        $ET_GUID_REF{ "GUID:$pb[1]" } and
        $ET_GUID_REF{ "IP:$pb[2]"   } and
        $ET_GUID_REF{ "GUID:$pb[1]" } eq $ET_GUID_REF{ "IP:$pb[2]" } )
      {
      $found = 1;
      last;
      }
    }
  return unless $found;

  my $cmd = $ET_GUID_REF_CMDS{ $arg[0] };
  et_rcon( $et_server, $cmd ) if $cmd;
}

#----------------------------------------------------------------------------

our %pl;        # players
our %ip;        # ip->name(s)
our %ip_rev;    # name->ip(s)
our %guid;      # guid->name(s)
our %guid_rev;  # name->guid(s)
our %same_ip;   # name->name(s) from the same ip
our %same_guid; # name->name(s) from the same guid

sub etresolve_load
{
  logger( undef, "ETRESOLVE: loading data..." );
  open( my $if, "/usr/bin/zcat $GUID_FILE |" );
  while( <$if> )
    {
    s/\^.//go;
    my $g = lc $1 if / ([0-9a-f]+)\(\-\)/i;
    my ( $i, $n ) = ( $1, lc $2 ) if / ([\d\.]+):\d+ +(.+)$/;
    
    $ip{ $i }{ $n }++;
    $ip_rev{ $n }{ $i }++;
    
    $guid{ $g }{ $n }++;
    $guid_rev{ $n }{ $g }++;
    
    $pl{ $n }++;
    }
  close( $if );
 
  while( my ( $n, $hr ) = each %ip_rev )
    {
    for my $i ( keys %$hr )
      {
      for my $nn ( keys %{ $ip{ $i } } )
        {
        $same_ip{ $n }{ $nn }++;
        }
      }
    }
  
  while( my ( $n, $hr ) = each %guid_rev )
    {
    for my $g ( keys %$hr )
      {
      next if $g eq 'unknown';
      for my $nn ( keys %{ $guid{ $g } } )
        {
        $same_guid{ $n }{ $nn }++;
        }
      }
    }
}

sub i_etresolve
{
  my $self = shift;
  my $ev   = shift;
  my $cmd  = lc shift;

  etresolve_load() unless %pl;
  etresolve_load() if $cmd eq '!reload';

  my @pl;
   
  if( $pl{ $cmd } )
    {
    push @pl, $cmd;
    }
  else  
    {
    @pl = grep /\S/, grep { index( $_, $cmd ) > -1 } keys %pl;
    
    if( @pl > 1 )
      {
      my $gi = $#pl < 9 ? $#pl : 9;
      return "*** !etr: several matches found: " . join( ' | ', @pl[0..$gi] ) . " ...";
      }
    if( @pl == 0 )
      {
      return "*** !etr: [$cmd] not found";
      }  
    }
    
  my @text;
  
  my $name = shift @pl;
  push @text, "*** !etr: found player [$name]";

  my $g = $same_guid{ $name };
  my $i = $same_ip{ $name };
  
  if( $g )
    {
    my $t;
    my $c = 1;
    for( sort keys %{ $g } )
      {
      $t .= "$_ | ";
      if( $c++ % 5 == 0 )
        {
        push @text, "same guid: $t";
        $t = '';
        }
      }
    }
  if( $i )
    {
    my $t;
    my $c = 1;
    for( sort keys %{ $i } )
      {
      $t .= "$_ | ";
      if( $c++ % 5 == 0 )
        {
        push @text, "same ip: $t";
        $t = '';
        }
      }
    }
 
  return join "\n", @text;
}

#############################################################################

sub et_stat
{
  my $server_nick = lc shift;
  
  my $server = $ETSTAT_SERVERS{ $server_nick };
  
  if( ! $server )
    {
    my $text = "*** unknown server [$server_nick] known servers are:\n";
    while( my ( $k, $v ) = each %ETSTAT_SERVERS )
      {
      $text .= "$k => $v\n";
      }
    return $text;
    }

  my $text;
  
  # FIXME: this should be replaced by the internal et_rcon()!!!
  open( my $i, "$QSTAT -P -R -raw '~' -rws $server |" );
  $_ = <$i>;
  chomp();
  # FIXME: this is replacment begining...
  #$_ = et_rcon( $server_nick, 'getinfo' );
  #print "ETSTATS: \n[$_]\n";
  my ( undef, $s_addr, $s_name, $map_name, $pl_max, $pl_cnt ) = split /~/;
  
  $_ = <$i>;
  chomp();
  my %env = split /[~=]/;
  
  my %side;
  
  if( $env{ 'Players_Allies' } and  $env{ 'Players_Axis'   } )
    {
    $side{ $_ } = 'Ally' for split /\s+/, $env{ 'Players_Allies' };
    $side{ $_ } = 'Axis' for split /\s+/, $env{ 'Players_Axis'   };
    }
  elsif( $env{ 'P' } )
    {
    my $c = 1;
    $side{ $c++ } = ( '????','Axis','Ally','Spct' )[$_] for split //, $env{ 'P' };
    }
  
  my %pl = ( 'Ally' => [], 'Axis' => [], 'Spct' => [] );

  my $c = 1;
  while(<$i>)
    {
    chomp();
    next unless /\S/;
    my ( $ping, $xp, $name ) = reverse split /~/;
    my $s = $side{ $c } || 'Spct';
    push @{ $pl{ $s } }, sprintf "%4s %-15s ", $xp, substr( $name, 0, 15 );
    $c++;
    }

  for my $s ( ( 'Axis', 'Ally', 'Spct' ) )
    {
    @{ $pl{ $s } } = reverse sort @{ $pl{ $s } };
    unshift @{ $pl{ $s } }, sprintf "%4s %-15s ", '-'x4, '-'x15;
    unshift @{ $pl{ $s } }, sprintf "%4s %-15s ", ' XP ', "$s Player";
    }
  
  $text .= "server: $s_name ($s_addr), $map_name ($pl_cnt/$pl_max)\n";
  return $text if $pl_cnt < 1;

  while(4)
    {
    $text .= shift @{ $pl{ $_ } } for( ( 'Axis', 'Ally', 'Spct' ) );
    $text .= "\n";
    last unless $c--;
    }
  $text =~ s/\n+$/\n/;
  $text .= '-'x62 . "\n";
  return $text;
}

#############################################################################

sub et_rcon
{
  my $server_nick = shift;
  my $command = shift;
  #sleep(1);
  
  my $server = $ETSTAT_SERVERS{ $server_nick } or 
      return "*** server [$server_nick] is unknown";
  my $pass = $ETRCON_PASSWD{ $server_nick } or
      return "*** server [$server_nick] is not available";
  
  $server .= ":27960" unless $server =~ /:\d+$/;
  my $sock = IO::Socket::INET->new( PeerAddr  => $server,
                                    Proto     => 'udp',
                                    LocalAddr => $HERE,
                                    Broadcast => '1' ) or 
  return "*** server [$server_nick] is temporary unavailable";
      
  my $sel = IO::Select->new();
  $sel->add($sock);
  
  if( $command eq 'getinfo' )
    {
    $sock->send( "\xff\xff\xff\xffgetinfo\n" );
    }
  else
    {  
    $sock->send( "\xff\xff\xff\xffrcon $pass $command" );
    }
  my $data;
  while( my @ready = $sel->can_read( 1 ) )
    {
    $sock->recv( $_, 4096 );
    s/\xff+(.*?)\n//;
    $data .= $_;
    };
  $data =~ s/\^.//g;
  
  return $data || "*** empty response";
}


###EOF#######################################################################
